@@ -38,11 +38,14 @@
 #include <stdio.h>
 #include <stdarg.h>
 #include <errno.h>
+#include <limits.h>
 
 #if USE_ZLIB
 # include <zlib.h>
 #endif
 
+#define FN_VMCOREINFO	"/sys/kernel/vmcoreinfo"
+
 /** Set an error message, returning @c kdump_status.
  * @arg err     Error message object.
  * @arg status  Status code.
@@ -74,6 +77,8 @@ kdump_err(kdump_ctx_t *ctx, kdump_status status, const char *msgfmt, ...)
 	if (status != KDUMP_OK) {
 		va_list ap;
 
+		if (status == KDUMP_ERR_SYSTEM && !err_str(&ctx->err))
+			err_add(&ctx->err, "%s", strerror(errno));
 		va_start(ap, msgfmt);
 		err_vadd(&ctx->err, msgfmt, ap);
 		va_end(ap);
@@ -154,7 +159,7 @@ mem_hash(const char *s, size_t len)
 	unsigned long hash = 0;
 
 	while (len >= sizeof(unsigned long)) {
-		add_to_hash(&hash, *(unsigned long*)s);
+		add_to_hash(&hash, get_unaligned_long((unsigned char *)s));
 		s += sizeof(unsigned long);
 		len -= sizeof(unsigned long);
 	}
@@ -163,12 +168,6 @@ mem_hash(const char *s, size_t len)
 	return hash;
 }
 
-unsigned long
-string_hash(const char *s)
-{
-	return mem_hash(s, strlen(s));
-}
-
 /**  Update a partial hash with a memory area.
  * @param[in,out] ph     Partial hash state.
  * @param[in]     s      Start of memory area with new data to be hashed.
@@ -179,23 +178,23 @@ phash_update(struct phash *ph, const char *s, size_t len)
 {
 	if (ph->idx) {
 		while (len && ph->idx < sizeof(unsigned long)) {
-			ph->part.bytes[ph->idx] = *s++;
+			ph->part[ph->idx] = *s++;
 			--len;
 			++ph->idx;
 		}
 		if (ph->idx >= sizeof(unsigned long)) {
-			add_to_hash(&ph->val, ph->part.num);
+			add_to_hash(&ph->val, get_unaligned_long(ph->part));
 			ph->idx = 0;
 		}
 	}
 
 	while (len >= sizeof(unsigned long)) {
-		add_to_hash(&ph->val, *(unsigned long*)s);
+		add_to_hash(&ph->val, get_unaligned_long((unsigned char*)s));
 		s += sizeof(unsigned long);
 		len -= sizeof(unsigned long);
 	}
 	while (len--)
-		ph->part.bytes[ph->idx++] = *s++;
+		ph->part[ph->idx++] = *s++;
 }
 
 static size_t
@@ -206,6 +205,7 @@ arch_ptr_size(enum kdump_arch arch)
 	case ARCH_ALPHA:
 	case ARCH_IA64:
 	case ARCH_PPC64:
+	case ARCH_RISCV64:
 	case ARCH_S390X:
 	case ARCH_X86_64:
 		return 8;	/* 64 bits */
@@ -214,6 +214,7 @@ arch_ptr_size(enum kdump_arch arch)
 	case ARCH_IA32:
 	case ARCH_MIPS:
 	case ARCH_PPC:
+	case ARCH_RISCV32:
 	case ARCH_S390:
 	default:
 		return 4;	/* 32 bits */
@@ -239,6 +240,10 @@ machine_arch_name(const char *machine)
 	else if (!strcmp(machine, "ppc64") ||
 		 !strcmp(machine, "ppc64le"))
 		return KDUMP_ARCH_PPC64;
+	else if (!strcmp(machine, "riscv32"))
+		return KDUMP_ARCH_RISCV32;
+	else if (!strcmp(machine, "riscv64"))
+		return KDUMP_ARCH_RISCV64;
 	else if (!strcmp(machine, "s390"))
 		return KDUMP_ARCH_S390;
 	else if (!strcmp(machine, "s390x"))
@@ -269,6 +274,8 @@ default_page_shift(enum kdump_arch arch)
 		[ARCH_MIPS] = 12,
 		[ARCH_PPC] = 0,
 		[ARCH_PPC64] = 0,
+		[ARCH_RISCV32] = 12,
+		[ARCH_RISCV64] = 12,
 		[ARCH_S390] = 12,
 		[ARCH_S390X] = 12,
 		[ARCH_IA32] = 12,
@@ -285,17 +292,19 @@ arch_ops(enum kdump_arch arch)
 {
 	switch (arch) {
 	case ARCH_ALPHA:
-	case ARCH_ARM:
 	case ARCH_IA64:
 	case ARCH_MIPS:
 	case ARCH_PPC:
+	case ARCH_RISCV32:
 	case ARCH_S390:
 		/* TODO */
 		break;
 
 	case ARCH_AARCH64:	return &aarch64_ops;
+	case ARCH_ARM:		return &arm_ops;
 	case ARCH_IA32:		return &ia32_ops;
 	case ARCH_PPC64:	return &ppc64_ops;
+	case ARCH_RISCV64:	return &riscv64_ops;
 	case ARCH_S390X:	return &s390x_ops;
 	case ARCH_X86_64:	return &x86_64_ops;
 
@@ -320,6 +329,8 @@ static const char canon_arch_names[][MAX_ARCH_NAME_LEN] =
 	DEF_ARCH(MIPS),
 	DEF_ARCH(PPC),
 	DEF_ARCH(PPC64),
+	DEF_ARCH(RISCV32),
+	DEF_ARCH(RISCV64),
 	DEF_ARCH(S390),
 	DEF_ARCH(S390X),
 	DEF_ARCH(X86_64),
@@ -664,15 +675,18 @@ set_uts(kdump_ctx_t *ctx, const struct new_utsname *src)
 {
 	static const struct {
 		unsigned idx, off;
+		bool empty_ok;
 	} defs[] = {
-#define DEF(name) \
-		{ GKI_linux_uts_ ## name, offsetof(struct new_utsname, name) }
-		DEF(sysname),
-		DEF(nodename),
-		DEF(release),
-		DEF(version),
-		DEF(machine),
-		DEF(domainname)
+#define DEF(name, empty_ok)						\
+		{ GKI_linux_uts_ ## name,				\
+		  offsetof(struct new_utsname, name),			\
+		  empty_ok }
+		DEF(sysname, false),
+		DEF(nodename, true),
+		DEF(release, false),
+		DEF(version, false),
+		DEF(machine, false),
+		DEF(domainname, true)
 #undef DEF
 	};
 
@@ -682,7 +696,7 @@ set_uts(kdump_ctx_t *ctx, const struct new_utsname *src)
 	for (i = 0; i < ARRAY_SIZE(defs); ++i) {
 		struct attr_data *d = gattr(ctx, defs[i].idx);
 		const char *s = (const char*)src + defs[i].off;
-		if (*s || !attr_isset(d)) {
+		if (*s || defs[i].empty_ok) {
 			res = set_uts_string(ctx, d, s);
 			if (res != KDUMP_OK)
 				return res;
@@ -855,23 +869,15 @@ cksum32(void *buffer, size_t size, uint32_t csum)
 kdump_status
 get_symbol_val(kdump_ctx_t *ctx, const char *name, kdump_addr_t *val)
 {
-	addrxlat_sym_t sym;
-	addrxlat_cb_t *cb;
+	const addrxlat_cb_t *cb;
 	addrxlat_status status;
 
-	cb = addrxlat_ctx_get_ecb(ctx->xlatctx);
-	if (!cb->sym)
-		return set_error(ctx, KDUMP_ERR_NODATA, "NULL callback");
-
-	sym.type = ADDRXLAT_SYM_VALUE;
-	sym.args[0] = name;
-	status = cb->sym(cb->data, &sym);
-	if (status != ADDRXLAT_OK)
-		return set_error(ctx, addrxlat2kdump(ctx, status),
-				 "Cannot resolve \"%s\"", sym.args[0]);
-
-	*val = sym.val;
-	return KDUMP_OK;
+	cb = addrxlat_ctx_get_cb(ctx->xlatctx);
+	status = cb->sym_value(cb, name, val);
+	return status != ADDRXLAT_OK
+		? set_error(ctx, addrxlat2kdump(ctx, status),
+			    "Cannot resolve \"%s\"", name)
+		: KDUMP_OK;
 }
 
 /**  Get the CPU directory attribute.
@@ -981,6 +987,26 @@ init_cpu_prstatus(kdump_ctx_t *ctx, unsigned cpu,
 	return init_cpu_blob_attr(ctx, cpu, data, size, &prstatus_tmpl);
 }
 
+/**  QEMU_CPUSTATE blob attribute template. */
+static const struct attr_template qemu_cpustate_tmpl = {
+	.key = "QEMU_CPUSTATE",
+	.type = KDUMP_BLOB,
+};
+
+/**  Initialize the QEMU_CPUSTATE attribute for a CPU
+ * @param ctx   Dump object.
+ * @param cpu   CPU number.
+ * @param data  QEMUCPUState raw binary data.
+ * @param size  Size of the QEMUCPUState data in bytes.
+ * @returns     Error status.
+ */
+kdump_status
+init_qemu_cpustate(kdump_ctx_t *ctx, unsigned cpu,
+		   const void *data, size_t size)
+{
+	return init_cpu_blob_attr(ctx, cpu, data, size, &qemu_cpustate_tmpl);
+}
+
 /**  XEN_PRSTATUS blob attribute template. */
 static const struct attr_template xen_prstatus_tmpl = {
 	.key = "XEN_PRSTATUS",
@@ -1042,7 +1068,7 @@ derived_attr_revalidate(kdump_ctx_t *ctx, struct attr_data *attr,
 			const struct attr_template *tmpl)
 {
 	const struct derived_attr_def *def = attr_to_derived_def(attr);
-	kdump_blob_t *blob;
+	kdump_blob_t *blob = NULL;
 	kdump_status status;
 	void *ptr;
 
@@ -1096,7 +1122,7 @@ derived_attr_update(kdump_ctx_t *ctx, struct attr_data *attr,
 		    const struct attr_template *tmpl)
 {
 	const struct derived_attr_def *def = attr_to_derived_def(attr);
-	kdump_blob_t *blob;
+	kdump_blob_t *blob = NULL;
 	kdump_status status;
 	void *ptr;
 
@@ -1132,7 +1158,7 @@ derived_attr_update(kdump_ctx_t *ctx, struct attr_data *attr,
 				   "Writing %hu-byte values not implemented",
 				   def->length);
 	}
-	attr->flags.invalid = 1;
+	attr->flags.invalid = true;
 
  unpin:
 	internal_blob_unpin(blob);
@@ -1159,8 +1185,10 @@ create_derived_attr(kdump_ctx_t *ctx, struct attr_data *dir,
 
 	action = "allocate";
 	attr = new_attr(ctx->dict, dir, &def->tmpl);
-	if (!attr)
+	if (!attr) {
+		status = KDUMP_ERR_SYSTEM;
 		goto err;
+	}
 
 	action = "set";
 	status = set_attr_number(ctx, attr, ATTR_INVALID, 0);
@@ -1230,6 +1258,61 @@ create_cpu_regs(kdump_ctx_t *ctx, unsigned cpu,
 	return status;
 }
 
+/**  Get register value from the corresponding QEMU_CPUSTATE attribute.
+ * @param ctx   Dump object.
+ * @param attr  Register value attribute.
+ * @returns     Error status.
+ */
+static kdump_status
+qemu_cpustate_reg_revalidate(kdump_ctx_t *ctx, struct attr_data *attr)
+{
+	return derived_attr_revalidate(ctx, attr, &qemu_cpustate_tmpl);
+}
+
+/**  Update the QEMU_CPUSTATE attribute after setting the register value.
+ * @param ctx   Dump object.
+ * @param attr  Register value attribute.
+ * @returns     Error status.
+ */
+static kdump_status
+qemu_cpustate_reg_update(kdump_ctx_t *ctx, struct attr_data *attr)
+{
+	return derived_attr_update(ctx, attr, &qemu_cpustate_tmpl);
+}
+
+/**  Create a set of CPU register attributes.
+ * @param ctx   Dump object.
+ * @param cpu   CPU number.
+ * @param def   Register definitions (array).
+ * @param ndef  Number of entries in the @c def array.
+ * @returns     Error status.
+ *
+ * Use this function to create register attributes under
+ * cpu.<num>.reg.  The values are taken from the corresponding
+ * cpu.<num>.QEMU_CPUSTATE blob attribute, using offsets and sizes
+ * from the definition array.
+ */
+kdump_status
+create_qemu_cpu_regs(kdump_ctx_t *ctx, unsigned cpu,
+		     struct derived_attr_def *def, unsigned ndef)
+{
+	static const struct attr_ops qemu_cpustate_reg_ops = {
+		.revalidate = qemu_cpustate_reg_revalidate,
+		.post_set = qemu_cpustate_reg_update,
+	};
+
+	struct attr_data *dir;
+	kdump_status status;
+
+	status = cpu_regs_dir(ctx, cpu, &dir);
+	while (status == KDUMP_OK && ndef--) {
+		def->tmpl.ops = &qemu_cpustate_reg_ops;
+		status = create_derived_attr(ctx, dir, def++);
+	}
+
+	return status;
+}
+
 /**  Get register value from the corresponding XEN_PRSTATUS attribute.
  * @param ctx   Dump object.
  * @param attr  Register value attribute.
@@ -1296,3 +1379,131 @@ set_file_description(kdump_ctx_t *ctx, const char *name)
 	return set_attr_static_string(ctx, gattr(ctx, GKI_file_description),
 				      ATTR_DEFAULT, name);
 }
+
+/** Set a blob attribute.
+ * @param ctx   Dump file object.
+ * @param attr  Global key idx.
+ * @param data  Raw data.
+ * @param size  Size of raw data.
+ * @param desc  Human-readable name of the blob (for error messages).
+ */
+kdump_status
+set_blob_attr(kdump_ctx_t *ctx, enum global_keyidx attr,
+	       void *data, size_t size, const char *desc)
+{
+	kdump_attr_value_t val;
+	kdump_status res;
+
+	val.blob = internal_blob_new_dup(data, size);
+	if (!val.blob)
+		return set_error(ctx, KDUMP_ERR_SYSTEM,
+				 "Cannot allocate %s", desc);
+	res = set_attr(ctx, gattr(ctx, attr), ATTR_DEFAULT, &val);
+	if (res != KDUMP_OK)
+		return set_error(ctx, res, "Cannot set %s", desc);
+
+	return KDUMP_OK;
+}
+
+/** Read a blob attribute from file.
+ * @param ctx   Dump file object.
+ * @param fidx  File index.
+ * @param off   File offset.
+ * @param size  Size of the blob.
+ * @param attr  Global attribute key index.
+ * @param desc  Human-readable name of the blob.
+ */
+kdump_status
+read_blob_attr(kdump_ctx_t *ctx, unsigned fidx, off_t off, size_t size,
+	       enum global_keyidx attr, const char *desc)
+{
+	struct fcache_chunk fch;
+	kdump_status ret;
+
+	ret = flatmap_get_chunk(ctx->shared->flatmap, &fch, size, fidx, off);
+	if (ret != KDUMP_OK)
+		return set_error(ctx, ret,
+				 "Cannot read %s (%zu bytes at %llu in %s)",
+				 desc, size, (unsigned long long) off,
+				 err_filename(ctx, fidx));
+
+	ret = set_blob_attr(ctx, attr, fch.data, size, desc);
+
+	fcache_put_chunk(&fch);
+	return ret;
+}
+
+/** Read running kernel's VMCOREINFO from
+ * @param ctx   Dump file object.
+ * @returns     Error status.
+ *
+ * Get VMCOREINFO location from /sys/kernel/vmcoreinfo and parse it
+ * using the currently opened dump file (which should be a live dump).
+ */
+kdump_status
+read_current_vmcoreinfo(kdump_ctx_t *ctx)
+{
+	FILE *f;
+	unsigned long long addr;
+	size_t length;
+	void *info;
+	kdump_status ret;
+
+	f = fopen(FN_VMCOREINFO, "r");
+	if (!f)
+		return errno == ENOENT
+			? KDUMP_OK
+			: set_error(ctx, KDUMP_ERR_SYSTEM,
+				    "Cannot open %s", FN_VMCOREINFO);
+
+	if (fscanf(f, "%llx %zx", &addr, &length) == 2)
+		ret = KDUMP_OK;
+	else if (ferror(f))
+		ret = set_error(ctx, KDUMP_ERR_SYSTEM,
+				"Error reading %s", FN_VMCOREINFO);
+	else
+		ret = set_error(ctx, KDUMP_ERR_CORRUPT,
+				"Error parsing %s: Wrong file format",
+				FN_VMCOREINFO);
+	fclose(f);
+	if (ret != KDUMP_OK)
+		return ret;
+
+	info = ctx_malloc(length, ctx, "VMCOREINFO buffer");
+	if (!info)
+		return KDUMP_ERR_SYSTEM;
+
+	ret = read_locked(ctx, KDUMP_MACHPHYSADDR, addr, info, &length);
+	if (ret == KDUMP_OK)
+		ret = process_notes(ctx, info, length);
+
+	free(info);
+	return ret;
+}
+
+/** Get file name for error messages.
+ * @param ctx   Dump file object.
+ * @param fidx  File index.
+ * @returns     Error status.
+ *
+ * The file name buffer is pre-allocated in @p ctx, so the returned
+ * string stays valid as long as @p ctx stays valid, or until the
+ * next call to this function.
+ */
+const char *
+err_filename(kdump_ctx_t *ctx, unsigned fidx)
+{
+	struct attr_data *attr;
+	const char *numstr;
+
+	sprintf(ctx->err_filename, "file #%u", fidx);
+	numstr = ctx->err_filename + sizeof("file #") - 1;
+	attr = lookup_dir_attr(ctx->dict, gattr(ctx, GKI_dir_file_set),
+			       numstr, strlen(numstr));
+	if (attr) {
+		attr = lookup_dir_attr(ctx->dict, attr, "name", 4);
+		if (attr && attr_isset(attr))
+			return attr_value(attr)->string;
+	}
+	return ctx->err_filename;
+}
